'Class:     Bitmap
'Author:    Todd Morrow
'Company:   Cosmologix
'Copyright: 1994-2010 Cosmologix
'Created:   10/26/2010 12:52:46 PM
'Purpose:   Bitmap operations

Public Class Bitmap

#Region "Constants"
#End Region

#Region "Enumerations"
  Public Enum PixelSizes
    Format32bppArgb = 4                  ' 4 bytes/pixel.
  End Enum
#End Region

#Region "Declares"
#End Region

#Region "Delegates"
#End Region

#Region "Structures"
#End Region

#Region "Events"
#End Region

#Region "Constructors"
  Public Sub New(ByVal Bitmap As System.Drawing.Bitmap)
    Me.Bitmap = Bitmap
  End Sub
#End Region

#Region "Properties"
  Private _Bytes As Byte()
  Public Property Bytes() As Byte()
    Get
      Return _Bytes
    End Get
    Set(ByVal Value As Byte())
      _Bytes = Value
    End Set
  End Property

  Private _Data As Imaging.BitmapData
  Private Property Data() As Imaging.BitmapData
    Get
      Return _Data
    End Get
    Set(ByVal Value As Imaging.BitmapData)
      _Data = Value
      If Data IsNot Nothing Then
        RowSize = Data.Stride
        ' Allocate room for the data.
        ReDim Bytes(Data.Stride * Data.Height - 1)
        ' Copy the data into the ImageBytes array.
        Marshal.Copy(Data.Scan0, Bytes, 0, Bytes.Length)
      Else
        Bytes = Nothing
        RowSize = 0
      End If
    End Set
  End Property

  Private _Bitmap As System.Drawing.Bitmap
  Public Property Bitmap() As System.Drawing.Bitmap
    Get
      Return _Bitmap
    End Get
    Private Set(ByVal Value As System.Drawing.Bitmap)
      _Bitmap = Value
      Select Case Bitmap.PixelFormat
        Case Imaging.PixelFormat.Format32bppArgb
          Me.PixelSize = PixelSizes.Format32bppArgb
        Case Else
          'TODO: Handle other kinds of bitmaps
          Throw New NotSupportedException("Invalid Bitmap PixelSize")
      End Select
      Width = Bitmap.Width
      Height = Bitmap.Height
      ReDim Coverage(Width - 1, Height - 1)
    End Set
  End Property

  Private _Height As Integer
  Public Property Height() As Integer
    Get
      Return _Height
    End Get
    Private Set(ByVal Value As Integer)
      _Height = Value
    End Set
  End Property

  Private _PixelSize As PixelSizes
  Public Property PixelSize() As PixelSizes
    Get
      Return _PixelSize
    End Get
    Private Set(ByVal Value As PixelSizes)
      _PixelSize = Value
    End Set
  End Property

  Private _RowSize As Integer
  Public Property RowSize() As Integer
    Get
      Return _RowSize
    End Get
    Private Set(ByVal Value As Integer)
      _RowSize = Value
    End Set
  End Property

  Private _Stack As Stack
  Public ReadOnly Property Stack() As Stack
    Get
      If _Stack Is Nothing Then
        _Stack = New Stack(1000)
      End If
      Return _Stack
    End Get
  End Property

  Private _Width As Integer
  Public Property Width() As Integer
    Get
      Return _Width
    End Get
    Private Set(ByVal Value As Integer)
      _Width = Value
    End Set
  End Property

  Private OldRed As Byte
  Private OldGreen As Byte
  Private OldBlue As Byte

  Private NewRed As Byte
  Private NewGreen As Byte
  Private NewBlue As Byte
  Private NewAlpha As Byte

  Private MinRed As Integer
  Private MinGreen As Integer
  Private MinBlue As Integer

  Private MaxRed As Integer
  Private MaxGreen As Integer
  Private MaxBlue As Integer

  Private Coverage(0, 0) As Boolean

#End Region

#Region "Methods"
  ''' <summary>
  ''' Flood the area at this point.
  ''' </summary>
  ''' <param name="X"></param>
  ''' <param name="Y"></param>
  ''' <param name="Color"></param>
  ''' <param name="Tolerance">
  ''' A number between 0 and 1 that represents the percent tolerance
  ''' </param>
  ''' <remarks></remarks>
  Public Sub FloodFill(ByVal X As Integer, ByVal Y As Integer, ByVal Color As Color, ByVal Tolerance As Single)

    Dim ToleranceLength As Integer
    'check / scale tolerance for byte comparisons
    If Tolerance > 1 OrElse Tolerance < 0 Then
      Throw New ArgumentException("Tolerance must be between 0 and 1")
    Else
      ToleranceLength = Math.Min(Tolerance * 255.0, 254)
    End If

    'TODO: ToleranceLength = 0 is a special case; simpler algorithm
    'check x,y arguments
    If X < 0 OrElse Y < 0 OrElse X > Width OrElse Y > Height Then
      Throw New ArgumentException("X and Y must be within bitmap bounds")
    End If

    'clear coverage area. Method might be called multiple times on same bitmap.
    ReDim Coverage(Width - 1, Height - 1)

    ' Get the old and new colors' components.
    With Bitmap.GetPixel(X, Y)
      OldRed = .R
      OldGreen = .G
      OldBlue = .B
    End With

    With Color
      NewRed = .R
      NewGreen = .G
      NewBlue = .B
      NewAlpha = .A
    End With

    'estabish tolerance comparison boundaries
    MinRed = OldRed - ToleranceLength
    MinGreen = OldGreen - ToleranceLength
    MinBlue = OldBlue - ToleranceLength

    MaxRed = OldRed + ToleranceLength
    MaxGreen = OldGreen + ToleranceLength
    MaxBlue = OldBlue + ToleranceLength

    ' Start with the original point in the stack.
    Stack.Push(New Point(X, Y))
    Bitmap.SetPixel(X, Y, Color)
    Coverage(X, Y) = True

    ' Lock the bitmap.
    Lock()

    ' While the stack is not empty, process a point.
    Dim Point As Point
    Do While Stack.Count > 0
      Point = Stack.Pop() 'is directcast faster?
      With Point
        'check four adjacent points to "flood" the fill area
        If .X > 0 Then CheckPoint(.X - 1, .Y)
        If .Y > 0 Then CheckPoint(.X, .Y - 1)
        If .X < _Width - 1 Then CheckPoint(.X + 1, .Y)
        If .Y < _Height - 1 Then CheckPoint(.X, .Y + 1)
      End With
    Loop

    ' Unlock the bitmap.
    Unlock()
  End Sub

  ' See if this point should be added to the stack.
  Private Sub CheckPoint(ByVal X As Integer, ByVal Y As Integer)

    'don't process same pixel twice
    If Coverage(X, Y) Then Exit Sub

    Dim Index As Integer = Y * _RowSize + X * _PixelSize
    Dim Blue As Byte = _Bytes(Index)
    Dim Green As Byte = _Bytes(Index + 1)
    Dim Red As Byte = _Bytes(Index + 2)

    'TODO: should alpha be used in comparison?
    If Red > MinRed AndAlso Red < MaxRed AndAlso _
      Green > MinGreen AndAlso Green < MaxGreen AndAlso _
      Blue > MinBlue AndAlso Blue < MaxBlue Then
      _Stack.Push(New Point(X, Y))
      _Bytes(Index) = NewBlue
      _Bytes(Index + 1) = NewGreen
      _Bytes(Index + 2) = NewRed
      _Bytes(Index + 3) = NewAlpha
    End If

    'mark pixel as covered
    Coverage(X, Y) = True

  End Sub

  ' Lock the bitmap's data.
  Public Sub Lock()
    ' Lock the bitmap data.
    Data = Bitmap.LockBits(New Rectangle(0, 0, Width, Height), Imaging.ImageLockMode.ReadWrite, Bitmap.PixelFormat)
  End Sub

  ' Copy the data back into the Bitmap
  ' and release resources.
  Public Sub Unlock()
    ' Copy the data back into the bitmap.
    Marshal.Copy(Bytes, 0, Data.Scan0, Bytes.Length)

    ' Unlock the bitmap.
    Bitmap.UnlockBits(Data)

    ' Release resources.
    Data = Nothing
  End Sub
#End Region

#Region "Event Handlers"
#End Region

#Region "Shared Properties"
#End Region

#Region "Shared Methods"
#End Region

#Region "Operators"
#End Region

End Class
